/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2019 IndiScale GmbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
'use strict';

/**
 * The tour module contains functionality for the demo tour through the
 * functionality of the caosdb web interface and some of caosdb's general
 * functionality.
 *
 * The content of a tour is loaded from the config file `tour.json`, which
 * defines chapters, sections and pages.  Each page corresponds to a popover
 * window with some content, which can be activated by hint buttons.
 *
 * Because it is easier to write for humans, this documentation will talk about
 * a corresponding yaml file instead, which needs to be converted to a json file
 * before usage.  Please also consult `tour.example.yaml` stored in the `doc`
 * folder.
 *
 * # Structure of the tour.yaml file #
 *
 * The top-level node "tour" contains a boolean `active` which specifies if the
 * tour shall be active initially and a node `elements` where the chapters etc.
 * are stored.  `elements` is a list of `page_set`s, which are the chapters of
 * the tour.
 *
 * ## `page_set` ##
 *
 * Each `page_set` (e.g. chapter or section) defines the following properties:
 *
 * - page_set :: Name of the page set.
 * - active :: If the page set is active initially.  For more than one active
 *             page set, only the last one wins.  Default = false
 * - persistent_state :: If the page set's state shall be saved in the
 *       sessionStorage.  Default = true
 * - deactivate_other :: If this page set is exclusive, i.e. if other page sets
 *       are deactivated upon activating this one.  Default = true
 * - id :: The id is used for referencing to this element.
 * - elements :: Contains elements of this page_set, either more page_sets or
 *     pages.
 *
 * ## `page` ##
 *
 * Pages are presented to the user via buttons, which show a popover window upon
 * click.  The buttons' position is defined relative to some other element,
 * their visibility may change by activity on other pages.  Pages share some
 * properties with page sets, some properties are new:
 *
 * - page :: Name, shown inside the page button.
 * - active :: Initial state, default = true
 * - id :: Same as for `page_set`.
 * - target :: Element selector for the element to which the page button should
 *             be attached.
 * - button_position :: Where the button is, relative to the target element.
 *       One of `top`, `bottom`, `left`, `right` or a combination thereof.
 * - button_size :: One of `small`, `medium`, `large`.  If not given, the button
 *       will have an automatic, not necessarily circular, size.
 * - title :: Title of the tour page.
 * - content :: The content of the page.  Rendered as Markdown.
 * - highlighters :: Makes HTML elements of the content highlight other
 *       elements.  Organized as key-value pairs, or rather ID-selector pairs.
 * - css :: Extra CSS attributes for the content frame box.
 * - activation / deactivation :: Specifies when this page button should be
 *       shown or hidden.  This is done by two sub-properties:
 *     - event :: Which kind of event triggers the (de)activation.  Typical
 *                events for pages are `open.tour.page`, `close.tour.page`.
 *     - target :: Selector at which element the event listener is being
 *       registered (if existent).
 *
 */
var ERROR = 1
var WARNING = 2
var INFO = 3
var DEBUG = 4
var TRACE = 5

var tour = new function() {

    ///////////////////////////////////////////////////////////////////////////
    //                         Improving a bit on jquery                     //
    ///////////////////////////////////////////////////////////////////////////

    // :Contains is case insensitive
    jQuery.expr[':'].Contains = function(a, i, m) {
        return jQuery(a).text().toUpperCase()
            .indexOf(m[3].toUpperCase()) >= 0;
    };

    this.PageSet = class {
        constructor(parent_set, config) {
            if(typeof parent_set === "undefined") {
                throw new Error("param `parent_set` must not be undefined");
            }
            if(typeof config === "undefined") {
                throw new Error("param `config` must not be undefined");
            }

            this.parent_set = parent_set;
            this.config = config;
            this._elements = new Array();
            this.active = false;

            for (const element of config.elements) {
                const next = tour.add_tour_element(element, this);
                this._elements.push(next);
            }

            // set some defaults
            if (typeof this.config.active === "undefined") {
                this.config.active = false;
            }
            if (typeof this.config.old_state_active === "undefined") {
                this.config.old_state_active = this.config.active;
            }
            if (typeof this.config.persistent_state === "undefined") {
                this.config.persistent_state = true;
            }
            if (typeof this.config.deactivate_other === "undefined") {
                this.config.deactivate_other = true;
            }
        }

        _activate_by_id(id) {
            this.parent_set._activate_by_id(id);
        }

        set old_state_active(value) {
            if(this.config.old_state_active != value) {
                this.config.old_state_active = value;
                this.update();
            }
        }

        get old_state_active() {
            return this.config.old_state_active;
        }

        update() {
            this.parent_set.update();
        }

        get name() {
            return this.config.page_set;
        }

        get full_name() {
            return this.parent_set.full_name + "." + this.name;
        }

        get elements() {
            return this._elements;
        }

        create_menu_entry() {
            var page_set_entry = $("<li class='list-group-item caosdb-f-tour-overview-entry caosdb-v-tour-overview-entry-pageset'/>");
            var link = $("<a class='btn btn-link'>" + this.name + "</a>")[0];
            this.menu_entry = link;
            link.addEventListener("click", () => {this.activate();});

            var elements_list = $('<ul class="list-group"/>');
            for (const element of this.elements) {
                const next = element.create_menu_entry();
                elements_list.append(next);
            }

            page_set_entry.append(link);
            page_set_entry.append(elements_list);
            return page_set_entry;
        }

        /**
         * Activate PageSet: Initialize activation of elements.
         */
        activate() {
            if (!this.active) {
                tour.info("PageSet.activate: tour element '" + this.full_name + "'.");

                this._activate();

                this._on_activation();

                // activation triggers the initialization of the children
                for (const element of this._elements) {
                    element.init_activation(this.config.persistent_state);
                }
            }
        }


        highlight_menu_entry() {
            $(this.menu_entry).toggleClass("caosdb-v-tour-menu-entry-highlight", true);
        }

        unhighlight_menu_entry() {
            $(this.menu_entry).toggleClass("caosdb-v-tour-menu-entry-highlight", false);
        }

        _activate() {
            if (!this.active) {
                tour.debug("PageSet._activate tour element '" + this.full_name + "'.");
                this.active = true;
                this.old_state_active = true;
                this.highlight_menu_entry();

                // activation propagates to the parents
                this.parent_set._activate()
                this.parent_set.deactivate_other(this);
            }
        }

        _on_activation() {
            let hiding = this.config.on_activation_hide;
            if (hiding) {
                console.log("Tour page set hiding:");
                hiding = tour.assert_array(hiding);
                hiding.forEach((selector) => {
                    console.log(selector);
                    $(selector).addClass("caosdb-f-tour-hidden");
                });
            }
        }

        deactivate_other(trigger) {
            if(this.config.deactivate_other) {
                for (const element of this.elements) {
                    if(element instanceof tour.PageSet && element !== trigger) {
                        element.deactivate();
                    }
                }
            }
        }

        /**
         * Initialize activation of PageSet.
         */
        init_activation(restore_old_state=false) {
            tour.debug("PageSet.init_activation tour element '" + this.full_name + "'.");
            if (restore_old_state) {
                if (this.old_state_active) {
                    this.activate();
                }
            } else if (this.config.active) {
                this.activate();
            }
        }

        deactivate() {
            if (this.active) {
                tour.info("PageSet.deactivate tour element '" + this.full_name + "'.");
                this.old_state_active = false;
                this._deactivate();
            }
        }

        _deactivate() {
            if (this.active) {
                tour.debug("PageSet._deactivate tour element '" + this.full_name + "'.");
                this.active = false;
                this.unhighlight_menu_entry();

                // deactivation propagates to the children
                for (const element of this._elements) {
                    element._deactivate();
                }
                // Unhide everything
                $(".caosdb-f-tour-hidden").removeClass("caosdb-f-tour-hidden");

            }
        }


    }

    this.Page = class {
        constructor(parent_set, config) {
            if(typeof parent_set === "undefined") {
                throw new Error("param `parent_set` must not be undefined");
            }
            if(typeof config === "undefined") {
                throw new Error("param `config` must not be undefined");
            }

            this.active = false;
            this.parent_set = parent_set;
            this.config = config;
            this.button = this._create_tour_button(config.page, config.content, config.title, config.id, config.button_size, config.css);

            this._activate_button(this.button, config.activation, config.deactivation);
            this._position_button(this.button, config.target, config.button_position);
            if (config.button_css) {
                for (let key in config.button_css) {
                    this.button.style.setProperty(key, config.button_css[key]);
                }
            }
            this._apply_highlighters(this.button, config.highlighters);
            this._apply_activation_links(this.button);
            this._apply_hiding(this.button);

            // set some defaults
            if (typeof this.config.active === "undefined") {
                this.config.active = true;
            }
            if (typeof this.config.old_state_active === "undefined") {
                this.config.old_state_active = this.config.active;
            }
        }

        _activate_by_id(id) {
            this.parent_set._activate_by_id(id);
        }

        create_menu_entry() {
            var entry = $("<li class='list-group-item caosdb-f-tour-overview-entry caosdb-v-tour-overview-entry-page' />");
            var link = $("<a class='btn btn-link'>" + this.name + "</a>")[0];
            link.addEventListener("click", () => {this.activate();});
            entry.append(link);
            return entry[0];
        }

        set old_state_active(value) {
            if(this.config.old_state_active != value) {
                this.config.old_state_active = value;
                this.update();
            }
        }

        get old_state_active() {
            return this.config.old_state_active;
        }

        update() {
            this.parent_set.update();
        }

        get name() {
            return this.config.page;
        }

        get full_name() {
            return this.parent_set.full_name + ":" + this.name;
        }

        init_activation(restore_old_state=false) {
            tour.debug("Page.init_activation button '" + this.full_name + "'.");
            if (restore_old_state) {
                if (this.old_state_active) {
                    this.activate();
                }
            } else if (this.config.active) {
                this.activate();
            }
        }

        activate() {
            if (!this.active) {
                tour.info("Page.activate button '" + this.full_name + "'.");
                this.old_state_active = true;
                this.active = true;

                // activation propagates to the parents
                this.parent_set._activate()

                $(this.button).show();
            }
        }

        deactivate() {
            if (this.active) {
                tour.info("Page.deactivate button '" + this.full_name + "'.");
                this.old_state_active = false;

                this._deactivate();
            }
        }


        _deactivate() {
            if (this.active) {
                tour.debug("Page._deactivate button '" + this.full_name + "'.");
                this.active = false;
                $(this.button).hide();
                $(this.button).popover("hide");
                $(this.button).toggleClass("caosdb-f-tour-open-page", false);
            }
        }

        /**
         * Hook the button to (de)activation events.
         */
        _activate_button(button, activation, deactivation) {
            $(button).hide();
            if (typeof activation !== "undefined") {
                activation = tour.assert_array(activation);
                activation.forEach(
                    (act) => {
                        if (act === null) { return; }
                        tour._post_init_cb.push(() => {
                            var target = $(act.target)[0];
                            if (target) {
                                target.addEventListener(
                                    act.event,
                                    (e) => {this.activate();}
                                );
                            }
                        });
                    }
                )
            }
            if (typeof deactivation !== "undefined" ) {
                deactivation = tour.assert_array(deactivation);
                deactivation.forEach(
                    (deact) => {
                        if (deact === null) { return; }
                        tour._post_init_cb.push(() => {
                            var target = $(deact.target)[0];
                            if (target) {
                                target.addEventListener(
                                    deact.event,
                                    (e) => {this.deactivate();}
                                );
                            }
                        });
                    }
                )
            }
        }


        _position_button(button, target, position) {
            var sel = $(target).first()
            if (typeof sel.css("position") === "undefined"
                || sel.css("position") === "static") {
                sel.css("position", "relative");
            }

            var wrapper = $('<div class="caosdb-f-tour-button-wrapper" style="display: block; position: relative; text-align: center; height: 0px;"></div>');
            wrapper.append(button);

            $(button).css("position", "absolute");

            switch (position) {
                case "top":
                    sel.prepend(wrapper);
                    $(button).css("top", - Math.abs($(button).outerHeight()) / 2);
                    //$(button).css("left", "50%");
                    break;
                case "top-right":
                    sel.prepend(wrapper);
                    $(button).css("top", - Math.abs($(button).outerHeight()) / 2);
                    $(button).css("right", - Math.abs($(button).outerWidth()) / 2);
                    break;
                case "right":
                    wrapper.css("top", "50%");
                    wrapper.css("right", sel.css("padding-right"));
                    wrapper.css("position", "absolute");
                    sel.prepend(wrapper);
                    $(button).css("right", - Math.abs($(button).outerWidth()) / 2);
                    $(button).css("top", - Math.abs($(button).outerHeight()) / 2);
                    break;
                case "bottom-right":
                    sel.append(wrapper);
                    $(button).css("bottom", - Math.abs($(button).outerHeight()) / 2);
                    $(button).css("right", - Math.abs($(button).outerWidth()) / 2);
                    break;
                case "bottom":
                    wrapper.css("top", "100%");
                    wrapper.css("left", "50%");
                    wrapper.css("position", "absolute");
                    sel.append(wrapper);
                    $(button).css("margin-top", "5px");
                    $(button).css("top", 0);
                    $(button).css("left", 0);
                    $(button).css("transform", "translate(-50%, 0)");
                    break;
                case "bottom-left":
                    sel.append(wrapper);
                    $(button).css("bottom", - Math.abs($(button).outerHeight()) / 2);
                    $(button).css("left", - Math.abs($(button).outerWidth()) / 2);
                    break;
                case "left":
                    wrapper.css("top", "50%");
                    wrapper.css("position", "absolute");
                    sel.prepend(wrapper);
                    $(button).css("margin-right", "5px");
                    $(button).css("top", 0);
                    $(button).css("right", 0);
                    $(button).css("transform", "translate(0, -50%)");
                    break;
                default:
                    // top-left
                    sel.prepend(wrapper);
                    $(button).css("top", - Math.abs($(button).outerHeight()) / 2);
                    $(button).css("left", - Math.abs($(button).outerWidth()) / 2);
                    break;
            }

        }


        _apply_highlighter(highlighter, highlightable) {
            $(highlighter).hover(
                ()=>{
                    highlightable.toggleClass("caosdb-v-tour-highlight", true);
                },
                ()=>{
                    highlightable.toggleClass("caosdb-v-tour-highlight", false);
                }
            );
        }

        /**
         * Hook click events for `data-tour-activate` <a> elements.
         */
        _apply_activation_links(button) {
            $(button).on('shown.bs.popover', (e) => {
                $('a[data-tour-activate]').each((idx, element) => {
                    element.addEventListener("click", () => {
                        this._activate_by_id($(element).data("tour-activate"));
                    });
                });
            });
        }

        _apply_hiding(button) {
            $(button).on('shown.bs.popover', (e) => {
                let hiding = this.config.on_activation_hide;
                if (hiding) {
                    console.log("Tour page hiding:");
                    hiding = tour.assert_array(hiding);
                    hiding.forEach((selector) => {
                        console.log(selector);
                        $(selector).addClass("caosdb-f-tour-hidden");
                    });
                }
                let unhiding = this.config.on_activation_unhide;
                if (unhiding) {
                    console.log("Tour page unhiding:");
                    unhiding = tour.assert_array(unhiding);
                    unhiding.forEach((selector) => {
                        console.log(selector);
                        $(selector).removeClass("caosdb-f-tour-hidden");
                    });
                }
            })
        }

        _apply_highlighters(button, highlighters) {
            if (typeof button === "undefined") {
                throw new Error("button was undefined");
            }
            if (!(button instanceof HTMLElement)) {
                throw new Error("button must instanciate HTMLElement");
            }
            if (!highlighters) {
                return;
            }
            const ids = Object.keys(highlighters)
            var highlightable = {}
            for (const id of ids) {
                highlightable[id] = $(highlighters[id]);
                if (id == "button") {
                    // special case, don't look for the highlighter in the content
                    this._apply_highlighter(button, highlightable[id]);
                } else {
                    $(button).on('shown.bs.popover', (e) => {
                        console.log("Highlighting:")
                        console.log(highlighters[id]);
                        console.log(highlightable[id]);
                        this._apply_highlighter("#" + id, highlightable[id]);
                    });
                }
            }
        }

        _create_tour_button(name, content, title, id, size, css={},
                            placement="auto") {
            if (typeof name === "undefined") {
                throw new Error("name was undefined");
            }
            if (typeof content === "undefined") {
                throw new Error("content was undefined");
            }

            var button = $('<button class="caosdb-v-tour-button">' + name + '</button>');

            var markdown_content = tour.markdown_to_html(content);
            tour.debug(markdown_content);

            // Apply custom style immediately at popover creation.
            // Alternatively, we could create a custom style element which sets
            // the style for "#" + id + " + .popover"

            if (!("max-width" in css)) {
                css["max-width"] = "120em";
            }
            if (!("width" in css)) {
                css["width"] = "50em";
            }
            // console.log(css)
            let popover_style = "";
            for (let key in css) {
                popover_style += key + ": " + css[key] + "; ";
            }
            // Bootstrap 3.x: .popover-content, 4.x: .popover-body
            let popover_template = '<div class="popover" role="tooltip" style="'
                + popover_style
                + '"><div class="arrow"></div><h3 class="popover-header"></h3>'
                + '<div class="popover-body popover-content"></div></div>';

            button.popover({
                title: title,
                content: markdown_content,
                placement: placement,
                html: true,
                sanitize: false,
                trigger: 'manual',
                template: popover_template,
            });

            button.on("click", function(e) {
                if($(this).hasClass("caosdb-f-tour-open-page")) {
                    this.dispatchEvent(tour.close_page_event);
                } else {
                    this.dispatchEvent(tour.open_page_event);
                }
                $(this).toggleClass("caosdb-f-tour-open-page");

                // clicks on the tour button should not trigger any other action
                e.preventDefault();
                e.stopPropagation();
            });

            button[0].addEventListener("close.tour.page", function() {
                button.popover("hide");
            });

            button[0].addEventListener("open.tour.page", function() {
                var p_id = button.popover("show")[0].getAttribute("aria-describedby");
                var popover = $("#" + p_id);

                // hide all other popovers
                $(".caosdb-v-tour-button").each(function(index) {
                    if (this != button[0]) {
                        $(this).popover("hide");
                    }
                });
            });

            if(id) {
                button.attr("id", id);
            }
            if(size)
                $(button).toggleClass(size, true);

            return button[0];
        }

    }

    this.config = undefined;
    this.close_page_event = new Event("close.tour.page");
    this.open_page_event = new Event("open.tour.page");

    this._post_init_cb = []

    this.post_init = function() {
        for (const fn of this._post_init_cb) {
            fn();
        }
    }

    /*
     * 0 - SILENT
     * 1 - ERROR
     * 2 - WARNING
     * 3 - INFO
     * 4 - DEBUG
     * 5 - TRACE
     */
    this.verbosity_level = INFO;

    /**
     * Initialize the tour.
     *
     * The `refresh` argument is currently only used interactively on the debugging console.
     */
    this.init = async function _in(refresh) {
        try {
            tour.debug("initializing tour module, refresh: " + refresh);
            if (refresh === true) {
                tour.info("Refreshing tour state.");
                localStorage.removeItem("tour_state");
            }
            await tour.load_tour();
            tour.post_init();
        } catch (error) {
            globalError(error);
        }
    }

    this.load_tour = async function _lt() {
        var config = undefined;
        try {
            var old_state = JSON.parse(localStorage.getItem("tour_state"));
            if (old_state) {
                config = {tour: old_state}
            };
        } catch (error) {
            if (error instanceof SyntaxError) {
                tour.warning("Parsing old tour state failed with SyntaxError. Old tour state: '" + localStorage.getItem("tour_state") + "'.");
            } else {
                globalError(error);
            }
        }
        if(!config || config.length == 0 || config.tour.length == 0) {
            tour.info("No old tour state in the localStorage.");
            // try to fetch 
            config = await load_config("tour.json");
            console.log("Loaded tour.json:");
            console.log(config);
        }
        if (!config || config.length == 0 || config.tour.length == 0) {
            localStorage.setItem("tour_state", "[]");
            tour.info("Tour config is empty.");
            return;
        }

        tour.configure(config.tour);
    }




    this.add_tour_element = function(element, parent_set) {
        if (element.page_set) {
            // it's a page_set
            return tour.add_tour_page_set(element, parent_set);
        } else {
            // it's a page
            return tour.add_tour_page(element, parent_set);
        }
    }


    this.add_tour_page_set = function(config, parent_set) {
        return new tour.PageSet(parent_set, config);
    }

    this.Tour = class {
        constructor(config) {
            this.full_name = "Tour";
            this.config = config;
            this.elements = new Array();
            this.leave_tour_button = $('<a class="btn btn-link">Leave Tour</a>');
            this.leave_tour_button.hide();
            this.reset_tour_button = $('<a class="btn btn-link">Reset Tour</a>');
            this.reset_tour_button.hide();
            this.active = false;
            for (const element of this.config.elements) {
                const next = tour.add_tour_element(element, this);
                this.elements.push(next);
            }

            // set some defaults
            this.config.active = !!this.config.active;
            if (typeof this.config.old_state_active === "undefined") {
                this.config.old_state_active = this.config.active;
            }
            if (typeof this.config.persistent_state === "undefined") {
                this.config.persistent_state = true;
            }
            if (typeof this.config.deactivate_other === "undefined") {
                this.config.deactivate_other = true;
            }

            this.update();
            this.panel = this.create_tour_overview_panel();
            this.init_activation(this.config.persistent_state);

            this._elements_by_id = {};
            this._index_elements(this._elements_by_id, this);

        }


        _index_elements(index, element) {
            if(element.elements) {
                for (const sub of element.elements) {
                    this._index_elements(index, sub);
                    if (sub.config.id) {
                        index[sub.config.id] = sub;
                    }
                }
            }
        }


        _activate_by_id(id) {
            var element = this._elements_by_id[id];
            if (element) {
                element.activate();
            }
        }

        set old_state_active(value) {
            if(this.config.old_state_active != value) {
                this.config.old_state_active = value;
                this.update();
            }
        }

        get old_state_active() {
            return this.config.old_state_active;
        }

        /**
         * Start tour activation.
         */
        init_activation(restore_old_state=false) {
            tour.debug("Tour.init_activation '" + this.full_name + "'.");
            if (restore_old_state) {
                if (this.old_state_active) {
                    this.activate();
                }
            } else if (this.config.active) {
                this.activate();
            }
        }

        deactivate() {
            tour.info("Tour.deactivate tour");
            this._deactivate();

            // deactivation propagates to the children
            for (const element of this.elements) {
                element._deactivate();
            }
        }

        /**
         * Reset the tour state.  Mainly useful for development and debugging.
         */
        async reset_tour() {
            // console.log("Resetting the tour");
            this.deactivate();
            localStorage.removeItem("tour_state");
            var config = await load_config("tour.json")
            if (!config || config.length == 0 || config.tour.length == 0) {
                localStorage.setItem("tour_state", "[]");
                tour.info("Tour config is empty.");
            } else {
                tour.configure(config.tour);
            }
            // A bit ugly, but the reset button will be removed or completely
            // changed for production anyways.
            location.reload();
        }

        _deactivate() {
            if(this.active) {
                this.set_tour_button_text("Start A Tour");
                tour.debug("Tour._deactivate tour");
                this.old_state_active = false;
                this.active = false;
                this.leave_tour_button.hide();
                this.reset_tour_button.hide();
                this.update();
            }
        }


        _activate() {
            if(!this.active) {
                this.set_tour_button_text("Tour");
                tour.debug("Tour._activate tour");
                this.old_state_active = true;
                this.active = true;
                this.leave_tour_button.show();
                this.reset_tour_button.show();
                this.update();
            }
        }

        /**
         * Activate tour: initialize activation of elements.
         */
        activate() {
            tour.info("Tour.activate tour");
            this._activate();

            for (const element of this.elements) {
                element.init_activation(true);
            }

        }

        deactivate_other(trigger) {
            if(this.config.deactivate_other) {
                for (const element of this.elements) {
                    if(element instanceof tour.PageSet && element !== trigger) {
                        element.deactivate();
                    }
                }
            }
        }

        set_tour_button_text(text) {
            $("button.caosdb-f-tour-button").text(text);
        }


        create_tour_overview_panel() {
            var panel = $('<div class="collapse" id="caosdb-f-tour-overview-panel" /></div>');
            var tour_overview = $('<ul class="list-inline caosdb-v-tour-overview"/>');
            for (const element of this.elements) {
                const next = element.create_menu_entry();
                tour_overview.append(next);
            }

            panel.append(tour_overview);

            this.leave_tour_button.on("click", () => {this.deactivate();});
            this.reset_tour_button.on("click", () => {this.reset_tour();});
            panel.append(this.leave_tour_button);
            panel.append(this.reset_tour_button);
            return panel[0];

        }

        update() {
            localStorage.setItem("tour_state", JSON.stringify(this.config));
        }


    }

    this.configure = function(config) {
        tour.info({"configure tour": config});

        // clean up old tour elements (after reload)
        $("#caosdb-f-tour-overview-panel").remove();
        $("#caosdb-navbar-tour").remove();
        $(".caosdb-f-tour-button-wrapper").remove();

        // new stuff following
        var tour_button = $(`
<li id="caosdb-navbar-tour">
  <button class="navbar-btn btn btn-link caosdb-f-tour-button" data-toggle="collapse" data-target="#caosdb-f-tour-overview-panel">
    Start A Tour
  </button>
</li>`);
        $('.caosdb-navbar').append(tour_button[0]);
        tour._instance = new tour.Tour(config);

        if (config.reload) {
            $(config.reload.target).each(function(index) {
                this.addEventListener(config.reload.event, (e) => {tour.configure(tour._instance.config);}, true);
            });
        }

        if(tour._instance.active) {
            tour._instance.set_tour_button_text("Tour");
        }
        $('#caosdb-query-panel').before(tour._instance.panel);
        // hide, when the mouse leaves the navbar
        $('nav.navbar').hover(undefined, ()=>{$(tour._instance.panel).collapse('hide');});

    }


    /**
     * Send a message of a certain level to the logger.
     *
     * @param {string} message - The message which is to be send to the logger.
     * @param {integer} [level=TRACE] - The level from 1 (Error) to 5 (Trace).
     */
    this.log = function(message, level=TRACE) {
        if (level <= this.verbosity_level) {
            var names = ["Error","Warning", "Info", "Debug", "Trace"];
            if (typeof message === "string" || message instanceof String) {
                console.log({level: names[level-1], logger: "tour", message: message});
            } else {
                console.log({level: names[level-1], logger: "tour", object: message});
            }
        }
    }

    /**
     * Send a debug message to the logger.
     *
     * @param {string} message - The message which is to be send to the logger.
     */
    this.debug = function(message) {
        this.log(message, DEBUG);
    }

    this.warning = function(message) {
        this.log(message, WARNING);
    }

    this.error = function(message) {
        this.log(message, ERROR);
    }

    this.info = function(message) {
        this.log(message, INFO);
    }

    this.add_tour_page = function(config, parent_set) {
        return new tour.Page(parent_set, config);
    }

    this.markdown_to_html = function(content) {
        let converter = new showdown.Converter();
        return converter.makeHtml(content.trim());
    }

    this.assert_array = function(content) {
        if (! Array.isArray(content)) {
                    return [content];
        }
        return content;
    }

    /**
     * Calls server-side script `scriptname`.
     */
    this.run_script = async function(scriptname) {
        try {
            const script_result = await connection.runScript(scriptname);
            const retcode = script_result.getElementsByTagName("script")[0].getAttribute("code");
            if (parseInt(retcode) > 0) {
                throw ("An error occurred during execution of the server-side script:\n"
                       + script_result.getElementsByTagName("script")[0].outerHTML);
            }
        } catch (e) {
            globalError(e);
        };
        return true;
    }


};


$(document).ready(tour.init);
