/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2020 IndiScale GmbH <info@indiscale.com>
 * Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */

'use strict';

var ext_revisions_test_suite = function ($, ext_revisions, QUnit, edit_mode) {

    var datamodel = ext_revisions._datamodel;

    QUnit.module("ext_revisions.js", {
        before: function (assert) {
            // setup before module
            this.original_update_entity = edit_mode.update_entity;
            this.original_insert = transaction.insertEntitiesXml;
            this.original_retrieve = transaction.retrieveEntityById;
            this.original_query = query;
            ext_revisions._logger.setLevel("trace");
        },
        beforeEach: function (assert) {
            // setup before each test
            datamodel.obsolete = "UNITTESTObsolete";
            datamodel.revisionOf = "UNITTESTRevisionOf";
        },
        afterEach: function (assert) {
            // teardown after each test
            query = this.original_query;
            edit_mode.update_entity = this.original_update_entity;
            transaction.insertEntitiesXml = this.original_insert;
            transaction.retrieveEntityById = this.original_retrieve;
        },
        after: function (assert) {
            // teardown after module
        }
    });

    QUnit.test("_make_revision_of_property", async function(assert) {
        var p = await ext_revisions._make_revision_of_property("1234");
        var editfield = $(p).find(".caosdb-property-edit-value");
        var value = $(editfield).find("select").first()[0].selectedOptions[0].value;
        assert.ok($(p).hasClass("caosdb-f-entity-property"), "is property");
        assert.equal(value, "1234", "has value 1234");
        assert.equal(getPropertyName(p), datamodel.revisionOf, "has revisionOf name");
        assert.equal(getPropertyDatatype(p), datamodel.obsolete, "has Obsolete datatype");
    });

    /**
     * This is a rather complete test, not a unit test.
     */
    QUnit.test("update calls update_entity through proxy", async function (assert) {
        var done = assert.async(3);
        var done_query = assert.async(2);
        var ent_element = $('<div data-entity-id="15"><div class="caosdb-properties"/></div>')[0];

        // mock server responses to several requests...
        var retrieve_fun = async function(id) {
            assert.equal(id, "15", "retrieve id 15");
            done();
            return $(`<Record id="15"><Parent name="ORIG_PARENT"/></Record>`)[0];
        }
        var insert_fun = async function(xml) {
            var rec = xml.firstElementChild.firstElementChild;
            assert.equal(rec.id, "-1", "insert with tmp id");
            assert.equal($(rec).find("Parent").attr("name"), datamodel.obsolete, "Obsolete Parent");
            xml.firstElementChild.firstElementChild.id = "2345";
            console.log(xml2str(xml));
            done();
            return xml;
        };
        var update_fun = async function(ent_element) {
            var prop = edit_mode.getProperties(ent_element)[0];
            assert.equal(prop.name, datamodel.revisionOf, "has revisionOf");
            assert.equal(prop.value, "2345", "revisionOf 2345");
            done();
        };
        var query_fun = async function(query) {
            assert.ok(query.startsWith("FIND") && ( query.endsWith(datamodel.obsolete) || query.endsWith(datamodel.revisionOf)), query);
            done_query(); // called twice
            return [$(`<div data-entity-name="${datamodel.revisionOf}" data-caosdb-id="3456"/>`)[0]];
        }

        // injecting the server mock-up responses.
        transaction.retrieveEntityById = retrieve_fun;
        transaction.insertEntitiesXml = insert_fun;
        edit_mode.update_entity = update_fun;
        query = query_fun;


        // actual tests
        assert.equal(update_fun, edit_mode.update_entity, "before init, the edit_mode.update_entity function has not been overridden.");

        // call init which checks the datamodel and overwrites the
        // edit_mode.update_entity function.
        await ext_revisions.init();
        assert.notEqual(update_fun, edit_mode.update_entity, "after init, the edit_mode.update_entity hab been overriden with a proxy calling the update_fun and the original function.");

        // call edit_mode.update_entity which calls the insert_fun and the
        // update_fun
        await edit_mode.update_entity(ent_element);
    });

}($, ext_revisions, QUnit, edit_mode);
